import os
import sys
import ssl
import time
import Queue
import errno
import signal
import socket
import threading
import logging
import logging.config
from functools import partial

import pyev
from sjrpc.core import RpcConnection, RpcError
from sjrpc.utils import threadless

from cloudcontrol.common.client.tags import get_tags, RootTagDB
from cloudcontrol.common.client.exc import PluginError, ConfigError


logger = logging.getLogger(__name__)


class RPCStartHandler(object):
    """Handles rpc connection and authentication to the remote cc-server.

    It performs the following steps (modelized as a DFA):
    | Name service resolution
    ->
    | TCP connection
    ->
    | SSL handshake
    ->
    | authentication to the cc-server
    ->
    | done
    """

    AUTH_TIMEOUT = 30.

    def __init__(self, loop):
        """
        :param loop: MainLoop instance
        """
        # internal DFA state handling
        self._states = {
            # FORMAT IS: STATE: (INIT, CLEANUP)
            # INIT is a pre state execution that is called to setup things like
            # watchers, threads
            # CLEANUP is a post state that cleans what is set in the INIT
            self.handle_name: (self.init_name, self.cleanup_name),
            self.handle_connect: (self.init_connect, self.cleanup_connect),
            self.handle_handshake: (self.init_handshake, self.cleanup_handshake),
            self.handle_authentication: (self.init_authentication,
                                         self.cleanup_authentication),
            self.handle_done: (None, None),
            self.handle_error: (None, None),  # obviously error state
        }
        self._current_state = None

        self.loop = loop

        # we don't declare all attributes in the constructor since some are only
        # specified to a single state
        self.sock = None
        self.rpc_con = None
        # id for async rpc call
        self.auth_id = None

    def _goto(self, where):
        """Method that is used to perform a state change, it will automatically
        call init/cleanup methods if needed.

        :param where: state where we wanna go or None
        """
        cleanup = self._states.get(self._current_state, (None, None))[1]
        if cleanup is not None:
            cleanup()
        self._current_state = where
        if where is not None:
            init = self._states[where][0]
            if init is not None:
                init()
            where()

    def init_name(self):
        self.thread = threading.Thread(target=self._handle_name)
        self.thread.daemon = True
        self.async_w = self.loop.evloop.async(self.handle_name_cb)

    def handle_name(self):
        self.async_w.start()
        self.thread.start()

    # Method run in the thread
    def _handle_name(self):
        while True:
            logger.debug('getaddrinfo')
            try:
                self.addr_info = iter(socket.getaddrinfo(
                    self.loop.config.server_host,
                    self.loop.config.server_port,
                    socket.AF_UNSPEC,
                    socket.SOCK_STREAM,
                ))
            except socket.gaierror as exc:
                logger.error(
                    'Error while resolving hostname for cc-server: %s',
                    exc.strerror,
                )
            except Exception:
                logger.exception(
                    'Unexpected error while resolving cc-server hostname')
            else:
                break

            time.sleep(5.)

        self.async_w.send()

    def handle_name_cb(self, watcher, revents):
        self._goto(self.handle_connect)

    def cleanup_name(self):
        del self.thread
        self.async_w.stop()
        del self.async_w

    def init_connect(self):
        # watch for non-blocking connection
        self.sock_write_w = None
        # tick is used for connect -> name transition to introduce some delay
        # between attempts
        self.tick_w = self.loop.evloop.timer(2., 0., self.handle_connect_error)

    def handle_connect_error(self, watcher, revents):
        self._goto(self.handle_name)

    def handle_connect(self):
        while True:
            try:
                (
                    family,
                    type_,
                    proto,
                    cannonname,
                    sockaddr,
                ) = self.current_addr = self.addr_info.next()
            except StopIteration:
                logger.error('Did try all addrinfo entries for cc-server without '
                             'success, will try to resolve hostname again')
                # we need to cleanup socket for the connect -> name transition
                # only, so we don't write the code in the cleanup method
                if self.sock is not None:
                    self.sock.close()
                self.tick_w.start()
                return

            try:
                self.sock = socket.socket(family, type_, proto)
            except EnvironmentError as exc:
                logger.error(
                    'Cannot create socket to connect to cc-server '
                    '(Errno %s: %s)', exc.errno, exc.strerror,
                )
                sys.exit(1)
                return

            try:
                self.sock.setblocking(0)
            except EnvironmentError as exc:
                logger.error(
                    'Cannot set socket in non blocking mode (Errno %s: %s),'
                    ' will exit', exc.errno, exc.strerror,
                )
                sys.exit(1)
                return

            error = self.sock.connect_ex(sockaddr)

            if error == errno.EINPROGRESS:  # most likely to happen
                self.sock_write_w = self.loop.evloop.io(self.sock, pyev.EV_WRITE,
                                                     self.handle_connected)
                self.sock_write_w.start()
                return
            elif error:
                logger.error(
                    'Error while trying to connect to cc-server using %s',
                    sockaddr,
                )
                continue
            else:
                # connection already succeeded, very unlikely as we're using TCP
                self._goto(self.handle_handshake)
                return

    def handle_connected(self, watcher=None, revents=None):
        # check that there is no error
        error = self.sock.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
        if error:
            logger.error(
                'Error while trying to connect to cc-server using %s,'
                ' (Errno %s: %s)', self.current_addr[-1],
                error, os.strerror(error),
            )
            self._goto(self.handle_error)
            return

        self._goto(self.handle_handshake)
        return

    def cleanup_connect(self):
        if self.sock_write_w is not None:
            self.sock_write_w.stop()
        del self.sock_write_w
        if self.tick_w.active:
            self.tick_w.stop()
        del self.tick_w

    def init_handshake(self):
        self.sock = ssl.wrap_socket(self.sock, do_handshake_on_connect=False,
                                    ssl_version=ssl.PROTOCOL_TLSv1)
        self.sock_read_w = self.loop.evloop.io(self.sock, pyev.EV_READ,
                                        self.handle_handshake)
        self.sock_write_w = self.loop.evloop.io(self.sock, pyev.EV_WRITE,
                                         self.handle_handshake)

    def handle_handshake(self, watcher=None, revents=None):
        assert self.sock is not None
        try:
            self.sock.do_handshake()
        except ssl.SSLError as exc:
            if exc.args[0] == ssl.SSL_ERROR_WANT_READ:
                self.sock_write_w.stop()
                self.sock_read_w.start()
            elif exc.args[0] == ssl.SSL_ERROR_WANT_WRITE:
                self.sock_read_w.stop()
                self.sock_write_w.start()
            else:
                logger.exception('SSL error:')
                self._goto(self.handle_error)
                return
        except socket.error as exc:
            logger.error('Socket error while doing handshake, (errno %s: %s)',
                         exc.errno, exc.strerror)
            self._goto(self.handle_error)
            return
        else:
            # create rpc_connection object
            self.rpc_con = RpcConnection(
                sock=self.sock,
                loop=self.loop.evloop,
                handler=self.loop.rpc_handler,
                on_disconnect=self.loop.restart_rpc_connection,
            )
            # since the socket is referenced by rpc_con, we don't need it
            # anymore
            self.sock = None
            self._goto(self.handle_authentication)
            return

    def cleanup_handshake(self):
        self.sock_read_w.stop()
        del self.sock_read_w
        self.sock_write_w.stop()
        del self.sock_write_w

    def init_authentication(self):
        self.timeout_w = self.loop.evloop.timer(
            self.AUTH_TIMEOUT, 0., self.handle_authentication_timeout)
        # since the sjrpc connection start in fallback mode, it could block so
        # we need to check this before attempting authentication;
        # as there is no reliable way to check when we can authenticate, we try
        # every 2 seconds
        self.tick_w = self.loop.evloop.timer(
            .5, 2., self.handle_authentication_tick)
        self.auth_id = None

    def handle_authentication(self):
        assert self.sock is None
        assert self.rpc_con is not None

        self.tick_w.start()

    def handle_authentication_tick(self, watcher=None, revents=None):
        # check is fallback mode on sjrpc is set otherwise our call would block
        # the loop
        if not self.rpc_con._event_fallback.is_set():
            logger.debug('Will try authentication again latter')
            return

        self.tick_w.stop()
        self.timeout_w.start()
        # try to authenticate
        try:
            self.auth_id = self.rpc_con.rpc.async_call_cb(
                self.handle_authentication_cb,
                'authentify',
                self.loop.config.server_user,
                self.loop.config.server_passwd,
            )
        except RpcError as exc:
            if exc.exception == 'RpcConnectionError':
                logger.error('Authentication failed: connection lost')
            else:
                logger.exception('Unexpected exception while authenticating')

            self._goto(self.handle_error)
            return

    def handle_authentication_cb(self, call_id, response=None, error=None):
        # RPC callback
        assert call_id == self.auth_id

        if error is not None:
            # we got an error
            logger.error('Error while authenticating with cc-server: %s("%s")',
                         error['exception'], error.get('message', ''))

            self._goto(self.handle_error)
            return

        self.loop.rpc_con = self.rpc_con
        # The rest is up to the subclass :)
        self.handle_authentication_response(response)

    def handle_authentication_response(self, response):
        """To be subclassed. Decides what to do about authentication response."""
        raise NotImplementedError

    def handle_authentication_timeout(self, watcher, revents):
        logger.error('Timeout while authenticating with cc-server %s',
                     self.current_addr[-1])
        self._goto(self.handle_error)

    def cleanup_authentication(self):
        if self.tick_w.active:
            self.tick_w.stop()
        del self.tick_w
        if self.timeout_w.active:
            self.timeout_w.stop()
        del self.timeout_w

    def handle_done(self):
        logger.info('Successfully authenticated with role %s', str(self.loop.role))
        self.stop()

    def handle_error(self):
        # cleanup and goto handle_connect
        if self.rpc_con is not None:
            self.rpc_con.rpc._on_disconnect = None  # we don't want the cb
            self.rpc_con.shutdown()
            self.rpc_con = None
        elif self.sock is not None:
            self.sock.close()
            self.sock = None
        self._goto(self.handle_connect)
        return

    def start(self):
        self._goto(self.handle_name)
        return

    def stop(self):
        self._goto(None)
        return


class MainLoop(object):

    # DEFAULTS_TAGS = tuple()
    # CONNECT_CLASS = RPCStartHandler
    # CONFIG_CLASS = None

    def __init__(self, config_path):
        self.config_path = config_path
        # load config variables
        try:
            self.load_config()
        except ConfigError as exc:
            sys.exit(exc.message)
        # configure logging
        self.configure_logging()

        # always run evloop as debug, we don't want exceptions to be mute
        self.evloop = pyev.default_loop(flags=pyev.EVFLAG_SIGNALFD,
                                        debug=True)

        # keeps track of libev loop thread, create threading queue and an async
        # watcher for performing work in libev thread from external thread
        # see decorators in utils.py
        self.pyev_thread = threading.current_thread()
        self.async_queue = Queue.Queue()
        self.async_watcher = self.evloop.async(self.async_work_cb)

        # set signal watchers
        self.signals = {
            signal.SIGINT: self.stop,
            signal.SIGTERM: self.stop,
            signal.SIGUSR1: self.reload,
        }

        # turn into real watchers
        self.signals = dict((
            signal_,
            self.evloop.signal(signal_, cb),
        ) for signal_, cb in self.signals.iteritems())

        # rpc connection
        self.rpc_con = None

        self.connect = self.CONNECT_CLASS(self)
        self.reconnect = None

        # role
        self.role = None
        self.main_plugin = None

        # tag database
        self.tag_db = RootTagDB(self, tags=self.DEFAULT_TAGS)

        # handlers
        self.rpc_handler = dict(
            get_tags=partial(threadless(get_tags), self.tag_db['__main__']),
            sub_tags=self.sub_tags,
        )

        # plugins
        self.registered_plugins = set()

    @property
    def rpc_connected(self):
        return self.rpc_con is not None

    @property
    def rpc_authenticated(self):
        return self.rpc_connected and self.role is not None

    def load_config(self):
        self.config = self.CONFIG_CLASS(self.config_path)

    def configure_logging(self):
        raise NotImplementedError

    def call_in_main_thread(self, func, *args, **kwargs):
        """Performs work in pyev thread.
        """
        if threading.current_thread() is self.pyev_thread:
            # if called from main thread, don't do anything special
            return func(*args, **kwargs)
        else:
            class Return(object):
                def __init__(self):
                    self.return_value = None
                    self.exc = None
            return_ = Return()
            event = threading.Event()
            def cb():
                try:
                    return_.return_value = func(*args, **kwargs)
                except Exception as e:
                    return_.exc = e
                finally:
                    event.set()
            self.async_queue.put(cb)
            self.async_watcher.send()

            # wait for work to be processed and raise error
            event.wait()
            if return_.exc is not None:
                raise return_.exc
            return return_.return_value

    def async_work_cb(self, watcher, revents):
        logger.debug('Async work processing')
        while True:
            try:
                work = self.async_queue.get_nowait()
            except Queue.Empty:
                break
            try:
                work()
            except Exception:
                # this should not happen but just in case
                logger.exception('Error async work')

    # RPC handlers definitions
    @threadless
    def sub_tags(self, sub_id, tags=None):
        if sub_id == '__main__':
            # FIXME should we raise ?
            logger.debug('Invalid request for sub object')
            return {}

        sub_db = self.tag_db.get(sub_id)
        if sub_db is None:
            # FIXME should we also raise here ?
            logger.debug('Failed to find sub_id %s', sub_id)
            return {}

        return get_tags(sub_db, tags)

    @threadless
    def job_list(self):
        pass
    # End RPC handlers definitions

    def reset_handler(self, name, handl):
        self.rpc_handler[name] = handl

    def remove_handler(self, name):
        self.rpc_handler.pop(name, None)

    def register_plugin(self, plugin):
        # keep track of registered plugins
        if plugin in self.registered_plugins:
            raise PluginError('Plugin was already registered')
        self.registered_plugins.add(plugin)

        # register tags
        plugin.tag_db.set_parent(self.tag_db)

        # register handler
        self.rpc_handler.update(plugin.rpc_handler)

        plugin.start()

    def unregister_plugin(self, plugin):
        try:
            self.registered_plugins.remove(plugin)
        except KeyError:
            raise PluginError('Plugin was not registered, cannot remove')

        # remove tags
        plugin.tag_db.set_parent(None)

        # remove handlers
        for handler_name in plugin.rpc_handler:
            del self.rpc_handler[handler_name]

        plugin.stop()

    def close_plugins(self):
        """Unregister all plugins from the loop."""
        for plugin in self.registered_plugins.copy():
            self.unregister_plugin(plugin)

    def restart_rpc_connection(self, *args):
        if not self.rpc_connected:
            return

        # clear connection
        self.rpc_con = None

        logger.error('Lost connection to the cc-server, will attempt'
                     ' reconnection')

        # reconnection atempt in one second
        self.reconnect = self.evloop.timer(
            1., 0., self.restart_rpc_connection_cb)
        self.reconnect.start()

    def restart_rpc_connection_cb(self, *args):
        # attempt to connect to the cc-server again
        self.connect.start()
        self.reconnect.stop()
        self.reconnect = None

    def start(self):
        logger.info('Starting node')
        self.async_watcher.start()
        for signal_ in self.signals.itervalues():
            signal_.start()
        logger.debug('About to connect')
        self.connect.start()
        logger.debug('About to start ev_loop')
        try:
            self.evloop.start()
        except:
            logger.exception('Unexpected exception while running main loop')
            sys.exit(1)

    def stop(self, watcher=None, revents=None):
        logger.info('Exiting node...')
        if self.connect is not None:
            self.connect.stop()
        if self.reconnect is not None:
            self.reconnect.stop()
        # close rpc
        if self.rpc_con is not None:
            # disable callback to prevent trampoline calls
            self.rpc_con.rpc._on_disconnect = None
            self.rpc_con.shutdown()
        # close all plugins
        for plugin in self.registered_plugins:
            plugin.stop()
        self.registered_plugins = set()

        # FIXME check for closing of main tags that were not in plugin
        self.role = None
        self.main_plugin = None
        self.evloop.stop()

    def reload(self, watcher=None, revents=None):
        logger.info('Reloading logging configuration...')
        try:
            self.load_config()
            self.configure_logging()
        except Exception:
            logger.exception('Invalid config file')
